//
// Freely inspired from Graphics Gems IV (AP Professional)
//

#include "debug.h"

#include "vector2.h"
#include "vector3.h"
#include "matrix3.h"

matrix3_c::
matrix3_c (void)
{
}

matrix3_c::
matrix3_c (const vector_t d)
{
   V[0] = V[1] = V[2] = vector3_c(d);
}

matrix3_c::
matrix3_c (const vector3_c& v0, const vector3_c& v1, const vector3_c v2)
{
   V[0] = v0;
   V[1] = v1;
   V[2] = v2;
}

matrix3_c::
matrix3_c (const matrix3_c& m)
{
   V[0] = m.V[0];
   V[1] = m.V[1];
   V[2] = m.V[2];
}

matrix3_c& matrix3_c::
operator  = (const matrix3_c& m)
{
   V[0] = m.V[0];
   V[1] = m.V[1];
   V[2] = m.V[2];
   return *this;
}

matrix3_c& matrix3_c::
operator -= (const matrix3_c& m)
{
   V[0] -= m.V[0];
   V[1] -= m.V[1];
   V[2] -= m.V[2];
   return *this;
}

matrix3_c& matrix3_c::
operator += (const matrix3_c& m)
{
   V[0] += m.V[0];
   V[1] += m.V[1];
   V[2] += m.V[2];
   return *this;
}

matrix3_c& matrix3_c::
operator *= (const vector_t d)
{
   V[0] *= d;
   V[1] *= d;
   V[2] *= d;
   return *this;
}

matrix3_c& matrix3_c::
operator /= (const vector_t d)
{
   V[0] /= d;
   V[1] /= d;
   V[2] /= d;
   return *this;
}

vector3_c& matrix3_c::
operator [] (const word i)
{
   debug_if(i>VZ,"invalid index access.");
   return V[i];
}

matrix3_c  matrix3_c::
transpose (void)
{
   return matrix3_c
            (
               vector3_c(V[0][0],V[1][0],V[2][0]),
               vector3_c(V[0][1],V[1][1],V[2][1]),
               vector3_c(V[0][2],V[1][2],V[2][2])
            );
}

matrix3_c  matrix3_c::
inverse (void)
{
   matrix3_c   m1(*this);
   matrix3_c   m2(indentity2D());
   word        i, j, i1;

   for(j=0; j<3; j++)
   {
      i1 = j;
      for (i=j+1; i<3; i++)
      {
         if (fabs(m1.V[i].N[j]) > fabs (m1.V[i1].N[j]))
            i1 = i;
      }

      swap(m1.V[i1],m1.V[j]);
      swap(m2.V[i1],m2.V[j]);

      if (m1.V[j].N[j] == 0.0)
         fatal("Can invert matrix 3x3 - singular one");

      m2.V[j] /= m1.V[j].N[j];
      m1.V[j] /= m1.V[j].N[j];

      for (i=0; i<3; i++)
      {
         if (i!=j)
         {
            m2.V[i] -= m1.V[i].N[j]*m2.V[j];
            m1.V[i] -= m1.V[i].N[j]*m1.V[j];
         }
      }
   }
   return m2;
}

matrix3_c& matrix3_c::
apply     (vector_function_ptr function)
{
   V[0].apply(function);
   V[1].apply(function);
   V[2].apply(function);
   return *this;
}

matrix3_c operator - (const matrix3_c& m)
{
   return matrix3_c(-m.V[0],-m.V[1],-m.V[2]);
}

matrix3_c operator - (const matrix3_c& m1, const matrix3_c& m2)
{
   return matrix3_c(m1.V[0]-m2.V[0],m1.V[1]-m2.V[1],m1.V[2]-m2.V[2]);
}

matrix3_c operator + (const matrix3_c& m1, const matrix3_c& m2)
{
   return matrix3_c(m1.V[0]+m2.V[0],m1.V[1]+m2.V[1],m1.V[2]+m2.V[2]);
}

matrix3_c operator * (const matrix3_c& m1, const matrix3_c& m2)
{
   #define ROWCOL(i,j)  m1.V[i].N[0]*m2.V[0].N[j] + \
                        m1.V[i].N[1]*m2.V[1].N[j] + \
                        m1.V[i].N[2]*m2.V[2].N[j]
   return matrix3_c
            (
               vector3_c(ROWCOL(0,0),ROWCOL(0,1),ROWCOL(0,2)),
               vector3_c(ROWCOL(1,0),ROWCOL(1,1),ROWCOL(1,2)),
               vector3_c(ROWCOL(2,0),ROWCOL(2,1),ROWCOL(2,2))
            );
   #undef ROWCOL
}

matrix3_c operator * (const vector_t d, const matrix3_c& m)
{
   return matrix3_c(m.V[0]*d,m.V[1]*d,m.V[2]*d);
}

matrix3_c operator * (const matrix3_c& m , const vector_t d)
{
   return d*m;
}

matrix3_c operator / (const matrix3_c& m, const vector_t d)
{
   return matrix3_c(m.V[0]/d,m.V[1]/d,m.V[2]/d);
}

bool operator == (const matrix3_c& m1, const matrix3_c& m2)
{
	return (m1.V[0]==m2.V[0]) && (m1.V[1]==m2.V[1]) && (m1.V[2]==m2.V[2]);
}

bool operator != (const matrix3_c& m1, const matrix3_c& m2)
{
   return !(m1==m2);
}


void swap (matrix3_c& m1, matrix3_c& m2)
{
   matrix3_c temp(m1);
   m1 = m2;
   m2 = temp;
}

matrix3_c indentity2D (void)
{
   return matrix3_c
            (
               vector3_c(1.0,0.0,0.0),
               vector3_c(0.0,1.0,0.0),
               vector3_c(0.0,0.0,1.0)
            );
}

matrix3_c translation2D (vector2_c& v)
{
   return matrix3_c
            (
               vector3_c(1.0,0.0,v[VX]),
               vector3_c(0.0,1.0,v[VY]),
               vector3_c(0.0,0.0,1.0)
            );
}

matrix3_c rotation2D (vector2_c& center, vector_t angle_deg)
{
   vector_t angle_rad = angle_deg*M_PI/180.0;
   vector_t c = cos(angle_rad);
   vector_t s = sin(angle_rad);
   vector_t t = 1-c;

   return matrix3_c
            (
               vector3_c(  c,  -s, center[VX]*t + center[VY]*s),
               vector3_c(  s,   c, center[VY]*t - center[VX]*s),
               vector3_c(0.0, 0.0, 1.0)
            );
}

matrix3_c scaling2D (vector2_c& scale_vector)
{
   return matrix3_c
            (
               vector3_c(scale_vector[VX],0.0,0.0),
               vector3_c(0.0,scale_vector[VY],0.0),
               vector3_c(0.0,0.0,1.0)
            );
}