//
// Freely inspired from Graphics Gems IV sources by Jean-Francois Doue
//                      OORT raytracer           by Nicholas Wilt.
//

#include "debug.h"

#include "matrix4.h"

matrix4_c::
matrix4_c (void)
{
}

matrix4_c::
matrix4_c (const vector_t d)
{
   V[0] = V[1] = V[2] = V[3] = vector4_c(d);
}

matrix4_c::
matrix4_c (const vector4_c& v0, const vector4_c& v1, const vector4_c v2, const vector4_c v3)
{
   V[0] = v0;
   V[1] = v1;
   V[2] = v2;
   V[3] = v3;
}

matrix4_c::
matrix4_c (const matrix4_c& m)
{
   V[0] = m.V[0];
   V[1] = m.V[1];
   V[2] = m.V[2];
   V[3] = m.V[3];
}

matrix4_c& matrix4_c::
operator  = (const matrix4_c& m)
{
   V[0] = m.V[0];
   V[1] = m.V[1];
   V[2] = m.V[2];
   V[3] = m.V[3];
   return *this;
}

matrix4_c& matrix4_c::
operator -= (const matrix4_c& m)
{
   V[0] -= m.V[0];
   V[1] -= m.V[1];
   V[2] -= m.V[2];
   V[3] -= m.V[3];
   return *this;
}

matrix4_c& matrix4_c::
operator += (const matrix4_c& m)
{
   V[0] += m.V[0];
   V[1] += m.V[1];
   V[2] += m.V[2];
   V[3] += m.V[3];
   return *this;
}

matrix4_c& matrix4_c::
operator *= (const vector_t d)
{
   V[0] *= d;
   V[1] *= d;
   V[2] *= d;
   V[3] *= d;
   return *this;
}

matrix4_c& matrix4_c::
operator /= (const vector_t d)
{
   V[0] /= d;
   V[1] /= d;
   V[2] /= d;
   V[3] /= d;
   return *this;
}

vector4_c& matrix4_c::
operator [] (const word i)
{
   debug_if(i>VW,"invalid index access.");
   return V[i];
}

matrix4_c  matrix4_c::
transpose (void)
{
   return matrix4_c
            (
               vector4_c(V[0][0],V[1][0],V[2][0],V[3][0]),
               vector4_c(V[0][1],V[1][1],V[2][1],V[3][1]),
               vector4_c(V[0][2],V[1][2],V[2][2],V[3][2]),
               vector4_c(V[0][3],V[1][3],V[2][3],V[3][3])
            );
}

matrix4_c  matrix4_c::
inverse (void)
{
   matrix4_c   m1(*this);
   matrix4_c   m2(indentity3D());
   word        i, j, i1;

   for(j=0; j<4; j++)
   {
      i1 = j;
      for (i=j+1; i<4; i++)
      {
         if (fabs(m1.V[i].N[j]) > fabs (m1.V[i1].N[j]))
            i1 = i;
      }

      swap(m1.V[i1],m1.V[j]);
      swap(m2.V[i1],m2.V[j]);

      if (m1.V[j].N[j] == 0.0)
         fatal("Can invert matrix 4x4 - singular one");

      m2.V[j] /= m1.V[j].N[j];
      m1.V[j] /= m1.V[j].N[j];

      for (i=0; i<4; i++)
      {
         if (i!=j)
         {
            m2.V[i] -= m1.V[i].N[j]*m2.V[j];
            m1.V[i] -= m1.V[i].N[j]*m1.V[j];
         }
      }
   }
   return m2;
}

matrix4_c& matrix4_c::
apply (vector_function_ptr function)
{
   V[0].apply(function);
   V[1].apply(function);
   V[2].apply(function);
   V[3].apply(function);
   return *this;
}

matrix4_c operator  - (const matrix4_c& m)
{
   return matrix4_c(-m.V[0],-m.V[1],-m.V[2],-m.V[3]);
}

matrix4_c operator  - (const matrix4_c& m1, const matrix4_c& m2)
{
   return matrix4_c(m1.V[0]-m2.V[0],m1.V[1]-m2.V[1],m1.V[2]-m2.V[2],m1.V[3]-m2.V[3]);
}

matrix4_c operator  + (const matrix4_c& m1, const matrix4_c& m2)
{
   return matrix4_c(m1.V[0]+m2.V[0],m1.V[1]+m2.V[1],m1.V[2]+m2.V[2],m1.V[3]+m2.V[3]);
}

matrix4_c operator  * (const matrix4_c& m1, const matrix4_c& m2)
{
   #define ROWCOL(i,j)  m1.V[i].N[0]*m2.V[0].N[j] + \
                        m1.V[i].N[1]*m2.V[1].N[j] + \
                        m1.V[i].N[2]*m2.V[2].N[j] + \
                        m1.V[i].N[3]*m2.V[3].N[j]
   return matrix4_c
            (
               vector4_c(ROWCOL(0,0),ROWCOL(0,1),ROWCOL(0,2),ROWCOL(0,3)),
               vector4_c(ROWCOL(1,0),ROWCOL(1,1),ROWCOL(1,2),ROWCOL(1,3)),
               vector4_c(ROWCOL(2,0),ROWCOL(2,1),ROWCOL(2,2),ROWCOL(2,3)),
               vector4_c(ROWCOL(3,0),ROWCOL(3,1),ROWCOL(3,2),ROWCOL(3,3))
            );
   #undef ROWCOL
}

matrix4_c operator  * (const vector_t d, const matrix4_c& m)
{
   return matrix4_c(m.V[0]*d,m.V[1]*d,m.V[2]*d,m.V[3]*d);
}

matrix4_c operator  * (const matrix4_c& m, const vector_t d)
{
   return d*m;
}

matrix4_c operator  / (const matrix4_c& m, const vector_t d)
{
   return matrix4_c(m.V[0]/d,m.V[1]/d,m.V[2]/d,m.V[3]/d);
}

bool operator == (const matrix4_c& m1, const matrix4_c& m2)
{
	return (m1.V[0]==m2.V[0]) && (m1.V[1]==m2.V[1]) && (m1.V[2]==m2.V[2]) && (m1.V[3]==m2.V[3]);
}

bool operator != (const matrix4_c& m1, const matrix4_c& m2)
{
   return !(m1==m2);
}

ostream& operator << (ostream& s, const matrix4_c& M)
{
   dword i,j;

   for (i=0; i<4; i++)
   {
      s << "| ";
      for (j=0; j<4; j++)
         s << M.V[i][j] << " ";
      s << ' |\n';
   }
   return s;
}


void swap (matrix4_c& m1, matrix4_c& m2)
{
   matrix4_c temp(m1);
   m1 = m2;
   m2 = temp;
}

matrix4_c indentity3D (void)
{
   return matrix4_c
            (
               vector4_c(1.0,0.0,0.0,0.0),
               vector4_c(0.0,1.0,0.0,0.0),
               vector4_c(0.0,0.0,1.0,0.0),
               vector4_c(0.0,0.0,0.0,1.0)
            );
}

matrix4_c translation3D (vector3_c& v)
{
   return matrix4_c
            (
               vector4_c(1.0,0.0,0.0,v[VX]),
               vector4_c(0.0,1.0,0.0,v[VY]),
               vector4_c(0.0,0.0,1.0,v[VZ]),
               vector4_c(0.0,0.0,0.0,1.0)
            );
}

matrix4_c rotation3D (vector3_c& axis, vector_t angle_deg)
{
   vector_t angle_rad = angle_deg*M_PI/180.0;
   vector_t c = cos(angle_rad);
   vector_t s = sin(angle_rad);
   vector_t t = 1-c;

   axis.normalize();

   return matrix4_c
            (
               vector4_c
                     (
                        t*axis[VX]*axis[VX] + c,
                        t*axis[VX]*axis[VY] - s*axis[VZ],
                        t*axis[VX]*axis[VZ] + c*axis[VY],
                        0.0
                     ),
               vector4_c
                     (
                        t*axis[VX]*axis[VY] + s*axis[VZ],
                        t*axis[VY]*axis[VY] + c,
                        t*axis[VY]*axis[VZ] - s*axis[VX],
                        0.0
                     ),
               vector4_c
                     (
                        t*axis[VX]*axis[VZ] - s*axis[VY],
                        t*axis[VY]*axis[VZ] + s*axis[VX],
                        t*axis[VZ]*axis[VZ] + c,
                        0.0
                     ),
               vector4_c
                     (
                        0.0, 0.0, 0.0, 1.0
                     )
            );
}

matrix4_c scaling3D (vector3_c& scale_vector)
{
   return matrix4_c
            (
               vector4_c(scale_vector[VX],0.0,0.0,0.0),
               vector4_c(0.0,scale_vector[VY],0.0,0.0),
               vector4_c(0.0,0.0,scale_vector[VZ],0.0),
               vector4_c(0.0,0.0,0.0,1.0)
            );
}

matrix4_c perspective3D (vector_t distance)
{
   return matrix4_c
            (
               vector4_c(distance,0.0,0.0,0.0),
               vector4_c(0.0,distance,0.0,0.0),
               vector4_c(0.0,0.0,0.0,0.0),
               vector4_c(0.0,0.0,-1.0,distance)
            );
}


// Invert the matrix using Gaussian elimination.  Not efficient,
// nor does it check for singular matrices.
/*
void
Matrix::Invert()
{
    int i, j, k;
    Matrix Out = IdentityMatrix();

    for (i = 0; i < 4; i++) {
	if (x[i][i] != 1.0) {
	    float divby = x[i][i];
	    for (j = 0; j < 4; j++) {
		Out.x[i][j] /= divby;
		x[i][j] /= divby;
	    }
	}
	for (j = 0; j < 4; j++) {
	    if (j != i) {
		if (x[j][i] != 0.0) {
		    float mulby = x[j][i];
		      for (k = 0; k < 4; k++) {
			  x[j][k] -= mulby * x[i][k];
			  Out.x[j][k] -= mulby * Out.x[i][k];
		      }
		}
	    }
	}
    }
    *this = Out;
}
*/

// Invert the given matrix using the above inversion routine.
/*
Matrix
Invert(const Matrix& M)
{
    Matrix InvertMe = M;
    InvertMe.Invert();
    return InvertMe;
}
*/

// Transpose the matrix.
/*
void
Matrix::Transpose()
{
    for (int i = 0; i < 4; i++)
	for (int j = i; j < 4; j++)
	    if (i != j) {
		float temp = x[i][j];
		x[i][j] = x[j][i];
		x[j][i] = temp;
	    }
}
*/

// Transpose the given matrix using the transpose routine above.
/*
Matrix
Transpose(const Matrix& M)
{
    Matrix TransposeMe = M;
    TransposeMe.Transpose();
    return TransposeMe;
}
*/

// Construct an identity matrix.
/*
Matrix
IdentityMatrix()
{
    Matrix M;

    for (int i = 0; i < 4; i++)
	for (int j = 0; j < 4; j++)
	    M.x[i][j] = (i == j) ? 1.0 : 0.0;
    return M;
}
*/

// Construct a zero matrix.
/*
Matrix
ZeroMatrix()
{
    Matrix M;
    for (int i = 0; i < 4; i++)
        for (int j = 0; j < 4; j++)
            M.x[i][j] = 0;
    return M;
}
*/

// Construct a translation matrix given the location to translate to.
/*
Matrix
TranslationMatrix(const Vector3D& Location)
{
    Matrix M = IdentityMatrix();
    M.x[3][0] = Location.x;
    M.x[3][1] = Location.y;
    M.x[3][2] = Location.z;
    return M;
}
*/

// Construct a rotation matrix.  Rotates Angle radians about the
// X axis.
/*
Matrix
RotationXMatrix(float Angle)
{
    Matrix M = IdentityMatrix();
    float Cosine = cos(Angle);
    float Sine = sin(Angle);
    M.x[1][1] = Cosine;
    M.x[2][1] = -Sine;
    M.x[1][2] = Sine;
    M.x[2][2] = Cosine;
    return M;
}
*/

// Construct a rotation matrix.  Rotates Angle radians about the
// Y axis.
/*
Matrix
RotationYMatrix(float Angle)
{
    Matrix M = IdentityMatrix();
    float Cosine = cos(Angle);
    float Sine = sin(Angle);
    M.x[0][0] = Cosine;
    M.x[2][0] = -Sine;
    M.x[0][2] = Sine;
    M.x[2][2] = Cosine;
    return M;
}
*/

// Construct a rotation matrix.  Rotates Angle radians about the
// Z axis.
/*
Matrix
RotationZMatrix(float Angle)
{
    Matrix M = IdentityMatrix();
    float Cosine = cos(Angle);
    float Sine = sin(Angle);
    M.x[0][0] = Cosine;
    M.x[1][0] = -Sine;
    M.x[0][1] = Sine;
    M.x[1][1] = Cosine;
    return M;
}
*/

// Construct a yaw-pitch-roll rotation matrix.	Rotate Yaw
// radians about the XY axis, rotate Pitch radians in the
// plane defined by the Yaw rotation, and rotate Roll radians
// about the axis defined by the previous two angles.
/*
Matrix
RotationYPRMatrix(float Yaw, float Pitch, float Roll)
{
    Matrix M;
    float ch = cos(Yaw);
    float sh = sin(Yaw);
    float cp = cos(Pitch);
    float sp = sin(Pitch);
    float cr = cos(Roll);
    float sr = sin(Roll);

    M.x[0][0] = ch*cr + sh*sp*sr;
    M.x[1][0] = -ch*sr + sh*sp*cr;
    M.x[2][0] = sh*cp;
    M.x[0][1] = sr*cp;
    M.x[1][1] = cr*cp;
    M.x[2][1] = -sp;
    M.x[0][2] = -sh*cr - ch*sp*sr;
    M.x[1][2] = sr*sh + ch*sp*cr;
    M.x[2][2] = ch*cp;
    for (int i = 0; i < 4; i++)
	M.x[3][i] = M.x[i][3] = 0;
    M.x[3][3] = 1;

    return M;
}
*/

// Construct a rotation of a given angle about a given axis.
// Derived from Eric Haines's SPD (Standard Procedural
// Database).
/*
Matrix
RotationAxisMatrix(const Vector3D& axis, float angle)
{
    Matrix M;
    double cosine = cos(angle);
    double sine = sin(angle);
    double one_minus_cosine = 1 - cosine;

    M.x[0][0] = axis.x * axis.x + (1.0 - axis.x * axis.x) * cosine;
    M.x[0][1] = axis.x * axis.y * one_minus_cosine + axis.z * sine;
    M.x[0][2] = axis.x * axis.z * one_minus_cosine - axis.y * sine;
    M.x[0][3] = 0;

    M.x[1][0] = axis.x * axis.y * one_minus_cosine - axis.z * sine;
    M.x[1][1] = axis.y * axis.y + (1.0 - axis.y * axis.y) * cosine;
    M.x[1][2] = axis.y * axis.z * one_minus_cosine + axis.x * sine;
    M.x[1][3] = 0;

    M.x[2][0] = axis.x * axis.z * one_minus_cosine + axis.y * sine;
    M.x[2][1] = axis.y * axis.z * one_minus_cosine - axis.x * sine;
    M.x[2][2] = axis.z * axis.z + (1.0 - axis.z * axis.z) * cosine;
    M.x[2][3] = 0;

    M.x[3][0] = 0;
    M.x[3][1] = 0;
    M.x[3][2] = 0;
    M.x[3][3] = 1;

    return M;
}
*/

// Construct a scale matrix given the X, Y, and Z parameters
// to scale by.  To scale uniformly, let X==Y==Z.
/*
Matrix
ScaleMatrix(float X, float Y, float Z)
{
    Matrix M = IdentityMatrix();

    M.x[0][0] = X;
    M.x[1][1] = Y;
    M.x[2][2] = Z;

    return M;
}
*/

// Construct a rotation matrix that makes the x, y, z axes
// correspond to the vectors given.
/*
Matrix
GenRotation(const Vector3D& x, const Vector3D& y, const Vector3D& z)
{
    Matrix M = IdentityMatrix();

    M.x[0][0] = x.x;
    M.x[1][0] = x.y;
    M.x[2][0] = x.z;
    M.x[0][1] = y.x;
    M.x[1][1] = y.y;
    M.x[2][1] = y.z;
    M.x[0][2] = z.x;
    M.x[1][2] = z.y;
    M.x[2][2] = z.z;

    return M;
}
*/

// Construct a view matrix to rotate and translate the world
// given that the viewer is located at Viewer, looking at
// LookAt, and the up vector is UpL.  The transformation
// changes things around so the viewer is at the origin and
// looking down the -Z axis.
/*
Matrix
ViewMatrix(const Vector3D& LookAt, const Vector3D& Viewer, const Vector3D& UpL)
{
    Matrix M = IdentityMatrix();
    Vector3D U, V, N;
    Vector3D Up = Normalize(UpL);

    N = Normalize(Viewer - LookAt);

    V = Up - LookAt;
    V -= N * DotProd(V, N);
    V = Normalize(V);
//    V = Normalize(Up - N * DotProd(Up, N));

    U = CrossProd(V, N);

    M.x[0][0] = U.x;
    M.x[1][0] = U.y;
    M.x[2][0] = U.z;
    M.x[0][1] = V.x;
    M.x[1][1] = V.y;
    M.x[2][1] = V.z;
    M.x[0][2] = N.x;
    M.x[1][2] = N.y;
    M.x[2][2] = N.z;

    M.x[3][0] = -DotProd(U, Viewer);
    M.x[3][1] = -DotProd(V, Viewer);
    M.x[3][2] = -DotProd(N, Viewer);

    return M;
}
*/

// Construct a quadric matrix.	After Foley et al. pp. 528-529.
/*
Matrix
QuadricMatrix(float a, float b, float c, float d, float e,
	      float f, float g, float h, float j, float k)
{
    Matrix M;

    M.x[0][0] = a;  M.x[0][1] = d;  M.x[0][2] = f;  M.x[0][3] = g;
    M.x[1][0] = d;  M.x[1][1] = b;  M.x[1][2] = e;  M.x[1][3] = h;
    M.x[2][0] = f;  M.x[2][1] = e;  M.x[2][2] = c;  M.x[2][3] = j;
    M.x[3][0] = g;  M.x[3][1] = h;  M.x[3][2] = j;  M.x[3][3] = k;

    return M;
}
*/

// Construct various "mirror" matrices, which flip coordinate
// signs in the various axes specified.
/*
Matrix
MirrorX()
{
    Matrix M = IdentityMatrix();
    M.x[0][0] = -1;
    return M;
}


Matrix
MirrorY()
{
    Matrix M = IdentityMatrix();
    M.x[1][1] = -1;
    return M;
}

Matrix
MirrorZ()
{
    Matrix M = IdentityMatrix();
    M.x[2][2] = -1;
    return M;
}

Matrix
RotationOnly(const Matrix& x)
{
    Matrix M = x;
    M.x[3][0] = M.x[3][1] = M.x[3][2] = 0;
    return M;
}
*/

// Apply the rotation portion of a matrix to a vector.
/*
Vector3D
RotateOnly(const Matrix& M, const Vector3D& v)
{
    Vector3D ret;
    float denom;

    ret.x = v.x * M.x[0][0] + v.y * M.x[1][0] + v.z * M.x[2][0];
    ret.y = v.x * M.x[0][1] + v.y * M.x[1][1] + v.z * M.x[2][1];
    ret.z = v.x * M.x[0][2] + v.y * M.x[1][2] + v.z * M.x[2][2];
    denom = M.x[0][3] + M.x[1][3] + M.x[2][3] + M.x[3][3];
    if (denom != 1.0)
	ret /= denom;
    return ret;
}
*/

// Scale each element of the matrix by B.
/*
Matrix
operator* (const Matrix& A, float B)
{
    Matrix ret;

    for (int i = 0; i < 4; i++)
	for (int j = 0; j < 4; j++)
	    ret.x[i][j] = A.x[i][j] * B;
    return ret;
}
*/

// Rotate a direction vector...
/*
Vector3D
PlaneRotate(const Matrix& tform, const Vector3D& p)
{
    // I sure hope that matrix is invertible...
    Matrix use = Transpose(Invert(tform));

    return RotateOnly(use, p);
}
*/