//
// Freely inspired from Graphics Gems IV (AP Professional)
//

#include "macro.h"
#include "debug.h"

#include "vector2.h"
#include "vector3.h"
#include "matrix3.h"

vector2_c::
vector2_c (void)
{
}

vector2_c::
vector2_c (const vector_t d)
{
   N[VX] = N[VY] = d;
}

vector2_c::
vector2_c (const vector_t x, const vector_t y)
{
   N[VX] = x;
	N[VY] = y;
}

vector2_c::
vector2_c (const vector2_c& v)
{
   N[VX] = v.N[VX];
   N[VY] = v.N[VY];
}

vector2_c::
vector2_c (const vector3_c& v)
{
	N[VX] = v.N[VX]/N[VZ];
   N[VY] = v.N[VY]/N[VZ];
}

vector2_c::
vector2_c (const vector3_c& v, word drop_axis)
{
   switch (drop_axis)
   {
      case VX: N[VX] = v.N[VY]; N[VY] = v.N[VZ]; break;
      case VY: N[VX] = v.N[VX]; N[VY] = v.N[VZ]; break;
      default: N[VX] = v.N[VX]; N[VY] = v.N[VY]; break;
   }
}

vector2_c& vector2_c::
operator  = (const vector2_c& v)
{
   N[VX] = v.N[VX];
   N[VY] = v.N[VY];
   return *this;
}

vector2_c& vector2_c::
operator -= (const vector2_c& v)
{
   N[VX] -= v.N[VX];
   N[VY] -= v.N[VY];
   return *this;
}

vector2_c& vector2_c::
operator += (const vector2_c& v)
{
   N[VX] += v.N[VX];
   N[VY] += v.N[VY];
   return *this;
}

vector2_c& vector2_c::
operator *= (const vector_t d)
{
   N[VX] *= d;
   N[VY] *= d;
   return *this;
}

vector2_c& vector2_c::
operator /= (const vector_t d)
{
   vector_t i = 1.0/d;
   debug_if(d==0, "division by zero");

   N[VX] *= i;
   N[VY] *= i;
   return *this;
}

vector2_c& vector2_c::
operator () (const vector_t x, const vector_t y)
{
   N[VX] = x;
   N[VY] = y;
   return *this;
}

vector2_c& vector2_c::
operator () (const vector2_c& v)
{
   N[VX] = v.N[VX];
   N[VY] = v.N[VY];
   return *this;
}

vector_t& vector2_c::
operator [] (const dword i)
{
   debug_if(i>VY,"illegal index access");
   return N[i];
}

vector_t vector2_c::
length  (void)
{
   return sqrt(length2());
}

vector_t vector2_c::
length2 (void)
{
   return N[VX]*N[VX]+N[VY]*N[VY];
}

vector2_c& vector2_c::
normalize (void)
{
   *this /= length();
   return *this;
}

vector2_c& vector2_c::
apply (vector_function_ptr function)
{
   N[VX] = (*function) (N[VX]);
   N[VY] = (*function) (N[VY]);
   return *this;
}

void vector2_c::
print (void)
{
   printf("(%4.4f, %4.4f)\n",N[VX],N[VY]);
}


void swap (vector2_c& v1, vector2_c& v2)
{
   vector2_c temp(v1);
   v1 = v2;
   v2 = temp;
}

vector2_c prod (const vector2_c& v1, const vector2_c& v2)
{
   return vector2_c(v1.N[VX]*v2.N[VX],v1.N[VY]*v2.N[VY]);
}

vector2_c min (const vector2_c& v1, const vector2_c& v2)
{
   return vector2_c(Min(v1.N[VX],v2.N[VX]),Min(v1.N[VY],v2.N[VY]));
}

vector2_c max (const vector2_c& v1, const vector2_c& v2)
{
   return vector2_c(Max(v1.N[VX],v2.N[VX]),Max(v1.N[VY],v2.N[VY]));
}

vector2_c operator- (const vector2_c& v)
{
   return vector2_c(-v.N[VX],-v.N[VY]);
}

vector2_c operator- (const vector2_c& v1, const vector2_c& v2)
{
   return vector2_c(v1.N[VX]-v2.N[VX],v1.N[VY]-v2.N[VY]);
}

vector2_c operator+ (const vector2_c& v1, const vector2_c& v2)
{
   return vector2_c(v1.N[VX]+v2.N[VX],v1.N[VY]+v2.N[VY]);
}

vector_t  operator* (const vector2_c& v1, const vector2_c& v2)
{
   return v1.N[VX]*v2.N[VX]+v1.N[VY]*v2.N[VY];
}

vector2_c operator* (const vector2_c& v, const vector_t d)
{
   return d*v;
}

vector2_c operator* (const vector_t d, const vector2_c& v)
{
   return vector2_c(d*v.N[VX],d*v.N[VY]);
}

vector2_c operator* (const vector2_c& v, matrix3_c& m)
{
   return m.transpose()*v;
}

vector2_c operator* (const matrix3_c& m, const vector2_c& v)
{
   vector3_c av;

   av.N[VX] = m.V[0].N[VX]*v.N[VX] + m.V[0].N[VY]*v.N[VY] + m.V[0].N[VZ];
   av.N[VY] = m.V[1].N[VX]*v.N[VX] + m.V[1].N[VY]*v.N[VY] + m.V[1].N[VZ];
   av.N[VZ] = m.V[2].N[VX]*v.N[VX] + m.V[2].N[VY]*v.N[VY] + m.V[2].N[VZ];
   // automatic casting
   return av;
}

vector2_c operator/ (const vector2_c& v, const vector_t d)
{
   vector_t i = 1.0/d;
   return vector2_c(i*v.N[VX],i*v.N[VY]);
}

vector3_c operator^ (const vector2_c& v1, const vector2_c& v2)
{
   return vector3_c(0.0,0.0,v1.N[VX]*v2.N[VY]-v2.N[VX]*v1.N[VY]);
}

bool operator == (const vector2_c& v1, const vector2_c& v2)
{
   return (v1.N[VX]==v2.N[VX]) && (v1.N[VY]==v2.N[VY]);
}

bool operator != (const vector2_c& v1, const vector2_c& v2)
{
   return !(v1==v2);
}

ostream& operator<< (ostream& s, const vector2_c& v)
{
    return s << "(" << v.N[VX] << ", " << v.N[VY] << ")";
}