//
// Freely inspired from Graphics Gems IV (AP Professional)
//

#include "debug.h"
#include "macro.h"

#include "vector2.h"
#include "vector3.h"
#include "vector4.h"
#include "matrix4.h"


vector3_c::
vector3_c (void)
{
}

vector3_c::
vector3_c (const vector_t d)
{
   N[VX] = N[VY] = N[VZ] = d;
}

vector3_c::
vector3_c (const vector_t x, const vector_t y, const vector_t z)
{
   N[VX] = x;
   N[VY] = y;
   N[VZ] = z;
}

vector3_c::
vector3_c (const vector3_c& v)
{
   N[VX] = v.N[VX];
   N[VY] = v.N[VY];
   N[VZ] = v.N[VZ];
}

vector3_c::
vector3_c (const vector2_c& v)
{
   N[VX] = v.N[VX];
   N[VY] = v.N[VY];
   N[VZ] = 0.0;
}

vector3_c::
vector3_c (const vector2_c& v, const vector_t d)
{
   N[VX] = v.N[VX];
   N[VY] = v.N[VY];
   N[VZ] = d;
}

vector3_c::
vector3_c (const vector4_c& v)
{
   N[VX] = v.N[VX]/v.N[VW];
   N[VY] = v.N[VY]/v.N[VW];
   N[VZ] = v.N[VZ]/v.N[VW];
}

vector3_c::
vector3_c (const vector4_c& v, word drop_axis)
{
   switch (drop_axis)
   {
      case VX: N[VX] = v.N[VY]; N[VY] = v.N[VZ]; N[VZ] = v.N[VW]; break;
      case VY: N[VX] = v.N[VX]; N[VY] = v.N[VZ]; N[VZ] = v.N[VW]; break;
      case VZ: N[VX] = v.N[VX]; N[VY] = v.N[VZ]; N[VZ] = v.N[VW]; break;
      default: N[VX] = v.N[VX]; N[VY] = v.N[VY]; N[VZ] = v.N[VZ]; break;
   }
}

vector3_c& vector3_c::
operator = (const vector3_c& v)
{
   N[VX] = v.N[VX];
   N[VY] = v.N[VY];
   N[VZ] = v.N[VZ];
   return *this;
}

vector3_c& vector3_c::
operator -= (const vector3_c& v)
{
   N[VX] -= v.N[VX];
   N[VY] -= v.N[VY];
   N[VZ] -= v.N[VZ];
   return *this;
}

vector3_c&
vector3_c:: operator += (const vector3_c& v)
{
   N[VX] += v.N[VX];
   N[VY] += v.N[VY];
   N[VZ] += v.N[VZ];
   return *this;
}

vector3_c& vector3_c::
operator *= (const vector_t d)
{
   N[VX] *= d;
   N[VY] *= d;
   N[VZ] *= d;
   return *this;
}

vector3_c& vector3_c::
operator /= (const vector_t d)
{
   vector_t i = 1.0/d;
   debug_if(d==0, "division by zero");

   N[VX] *= i;
   N[VY] *= i;
   N[VZ] *= i;
   return *this;
}

vector3_c& vector3_c::
operator () (const vector_t x, const vector_t y, const vector_t z)
{
   N[VX] = x;
   N[VY] = y;
   N[VZ] = z;
   return *this;
}

vector3_c& vector3_c::
operator () (const vector3_c& v)
{
	N[VX] = v.N[VX];
	N[VY] = v.N[VY];
	N[VZ] = v.N[VZ];
	return *this;
}


vector_t& vector3_c::
operator [] (const word i)
{
   debug_if(i>VZ,"illegal index access");
	return N[i];
}

vector_t vector3_c::
length  (void)
{
	return sqrt(length2());
}

vector_t vector3_c::
length2 (void)
{
   return N[VX]*N[VX]+N[VY]*N[VY]+N[VZ]*N[VZ];
}

vector3_c& vector3_c::
normalize (void)
{
   *this /= length();
   return *this;
}

vector3_c& vector3_c::
apply (vector_function_ptr function)
{
   N[VX] = (*function) (N[VX]);
   N[VY] = (*function) (N[VY]);
   N[VZ] = (*function) (N[VZ]);
   return *this;
}

void vector3_c::
print (void)
{
   printf("(%4.4f, %4.4f, %4.4f)",N[VX],N[VY],N[VZ]);
}

void swap (vector3_c& v1, vector3_c& v2)
{
   vector3_c temp(v1);
   v1 = v2;
	v2 = temp;
}

vector3_c prod (const vector3_c& v1,const vector3_c& v2)
{
   return vector3_c(v1.N[VX]*v2.N[VX],v1.N[VY]*v2.N[VY],v1.N[VZ]*v2.N[VZ]);
}

vector3_c min (const vector3_c& v1,const vector3_c& v2)
{
   return vector3_c(Min(v1.N[VX],v2.N[VX]),Min(v1.N[VY],v2.N[VY]),Min(v1.N[VZ],v2.N[VZ]));
}

vector3_c max (const vector3_c& v1,const vector3_c& v2)
{
   return vector3_c(Max(v1.N[VX],v2.N[VX]),Max(v1.N[VY],v2.N[VY]),Max(v1.N[VZ],v2.N[VZ]));
}

vector3_c operator - (const vector3_c& v)
{
   return vector3_c(-v.N[VX],-v.N[VY],-v.N[VZ]);
}

vector3_c operator - (const vector3_c& v1, const vector3_c& v2)
{
   return vector3_c(v1.N[VX]-v2.N[VX],v1.N[VY]-v2.N[VY],v1.N[VZ]-v2.N[VZ]);
}

vector3_c operator + (const vector3_c& v1, const vector3_c& v2)
{
   return vector3_c(v1.N[VX]+v2.N[VX],v1.N[VY]+v2.N[VY],v1.N[VZ]+v2.N[VZ]);
}

vector_t  operator * (const vector3_c& v1, const vector3_c& v2)
{
	return v1.N[VX]*v2.N[VX]+v1.N[VY]*v2.N[VY]+v1.N[VZ]*v2.N[VZ];
}

vector3_c operator * (const vector3_c& v, const vector_t d)
{
   return vector3_c(d*v.N[VX],d*v.N[VY],d*v.N[VZ]);
}

vector3_c operator * (const vector_t d, const vector3_c& v)
{
   return vector3_c(d*v.N[VX],d*v.N[VY],d*v.N[VZ]);
}

vector3_c operator * (const vector3_c& v, matrix4_c& m)
{
   return m.transpose()*vector4_c(v);
}

vector3_c operator * (const matrix4_c& m, const vector3_c& v)
{
   return m*vector4_c(v);
}

vector3_c operator / (const vector3_c& v, const vector_t d)
{
   vector_t i = 1.0/d;
   debug_if(d==0, "division by zero");

   return vector3_c(i*v.N[VX],i*v.N[VY],i*v.N[VZ]);
}

vector3_c operator ^ (const vector3_c& v1, const vector3_c& v2)
{
   return vector3_c
            (
               v1.N[VY]*v2.N[VZ] - v1.N[VZ]*v2.N[VY],
               v1.N[VZ]*v2.N[VX] - v1.N[VX]*v2.N[VZ],
               v1.N[VX]*v2.N[VY] - v1.N[VY]*v2.N[VX]
				);
}

bool operator == (const vector3_c& v1, const vector3_c& v2)
{
	return (v1.N[VX]==v2.N[VX]) && (v1.N[VY]==v2.N[VY]) && (v1.N[VZ]==v2.N[VZ]);
}

bool operator != (const vector3_c& v1, const vector3_c& v2)
{
	return !(v1==v2);
}

ostream& operator << (ostream& s, const vector3_c& v)
{
    return s << "(" << v.N[VX] << ", " << v.N[VY] << ", " << v.N[VZ] << ")";
}

vector_t cosinus (vector3_c& v1, vector3_c& v2)
{
	return (v1*v2)/(v1.length()*v2.length());
}
/*
vector_t  sinus (vector3_c& v1, vector3_c& v2)
{
	vector3_c v;
	v = v1^v2;
	return (v[VX]+v[VY]+v[VZ])/(v1.length()*v2.length());
}
*/

/*
void Minimize(Vector3D& min, const Vector3D& Candidate)
{
    if (Candidate.x < min.x)
	min.x = Candidate.x;
    if (Candidate.y < min.y)
	min.y = Candidate.y;
    if (Candidate.z < min.z)
	min.z = Candidate.z;
}

void Maximize(Vector3D& max, const Vector3D& Candidate)
{
    if (Candidate.x > max.x)
	max.x = Candidate.x;
	 if (Candidate.y > max.y)
	max.y = Candidate.y;
    if (Candidate.z > max.z)
	max.z = Candidate.z;
}
*/