//
// Freely inspired from Graphics Gems IV (AP Professional)
//

#include "debug.h"
#include "macro.h"

#include "vector3.h"
#include "vector4.h"
#include "matrix4.h"


vector4_c::
vector4_c (void)
{
}

vector4_c::
vector4_c (const vector_t d)
{
   N[VX] = N[VY] = N[VZ] = N[VW] = d;
}

vector4_c::
vector4_c (const vector_t x, const vector_t y, const vector_t z, const vector_t w)
{
   N[VX] = x;
   N[VY] = y;
   N[VZ] = z;
   N[VW] = w;
}

vector4_c::
vector4_c (const vector4_c& v)
{
   N[VX] = v.N[VX];
   N[VY] = v.N[VY];
   N[VZ] = v.N[VZ];
   N[VW] = v.N[VW];
}

vector4_c::
vector4_c (const vector3_c& v)
{
   N[VX] = v.N[VX];
   N[VY] = v.N[VY];
   N[VZ] = v.N[VZ];
   N[VW] = 1.0;
}

vector4_c::
vector4_c (const vector3_c& v, const vector_t d)
{
   N[VX] = v.N[VX];
   N[VY] = v.N[VY];
   N[VZ] = v.N[VZ];
   N[VW] = d;
}

vector4_c& vector4_c::
operator  = (const vector4_c& v)
{
   N[VX] = v.N[VX];
   N[VY] = v.N[VY];
   N[VZ] = v.N[VZ];
   N[VW] = v.N[VW];
   return *this;
}

vector4_c& vector4_c::
operator -= (const vector4_c& v)
{
   N[VX] -= v.N[VX];
   N[VY] -= v.N[VY];
   N[VZ] -= v.N[VZ];
   N[VW] -= v.N[VW];
   return *this;
}

vector4_c& vector4_c::
operator += (const vector4_c& v)
{
   N[VX] += v.N[VX];
   N[VY] += v.N[VY];
   N[VZ] += v.N[VZ];
   N[VW] += v.N[VW];
   return *this;
}

vector4_c& vector4_c::
operator *= (const vector_t d)
{
   N[VX] *= d;
   N[VY] *= d;
   N[VZ] *= d;
   N[VW] *= d;
   return *this;
}

vector4_c& vector4_c::
operator /= (const vector_t d)
{
   vector_t i = 1.0/d;
   debug_if(d==0, "division by zero");

   N[VX] *= i;
   N[VY] *= i;
   N[VZ] *= i;
   N[VW] *= i;
   return *this;
}

vector4_c& vector4_c::
operator () (const vector_t x, const vector_t y, const vector_t z, const vector_t w)
{
   N[VX] = x;
   N[VY] = y;
   N[VZ] = z;
   N[VW] = w;
   return *this;
}

vector4_c& vector4_c::
operator () (const vector4_c& v)
{
   N[VX] = v.N[VX];
   N[VY] = v.N[VY];
   N[VZ] = v.N[VZ];
   N[VW] = v.N[VW];
   return *this;
}

vector_t& vector4_c::
operator [] (const word i)
{
   debug_if(i>VW,"illegal index access");
   return N[i];
}

const vector_t& vector4_c::
operator [] (const word i) const
{
   debug_if(i>VW,"illegal index access");
   return N[i];
}

vector_t vector4_c::
length  (void)
{
   return sqrt(length2());
}

vector_t vector4_c::
length2 (void)
{
   return N[VX]*N[VX]+N[VY]*N[VY]+N[VZ]*N[VZ]+N[VW]*N[VW];
}

vector4_c& vector4_c::
normalize (void)
{
   *this /= length();
   return *this;
}

vector4_c& vector4_c::
homogene (void)
{
   N[VX] /= N[VW];
   N[VY] /= N[VW];
   N[VZ] /= N[VW];
   N[VW]  = 1.0;
   return *this;
}


vector4_c& vector4_c::
apply (vector_function_ptr function)
{
   N[VX] = (*function) (N[VX]);
   N[VY] = (*function) (N[VY]);
   N[VZ] = (*function) (N[VZ]);
   N[VW] = (*function) (N[VW]);
   return *this;
}

void vector4_c::
print (void)
{
   printf("(%4.4f, %4.4f, %4.4f, %4.4f)\n",N[VX],N[VY],N[VZ],N[VW]);
}

void swap (vector4_c& v1, vector4_c& v2)
{
   vector4_c temp(v1);
   v1 = v2;
   v2 = temp;
}

vector4_c prod (const vector4_c& v1,const vector4_c& v2)
{
   return vector4_c(v1.N[VX]*v2.N[VX],v1.N[VY]*v2.N[VY],v1.N[VZ]*v2.N[VZ],v1.N[VW]*v2.N[VW]);
}

vector4_c min (const vector4_c& v1,const vector4_c& v2)
{
   return vector4_c(Min(v1.N[VX],v2.N[VX]),Min(v1.N[VY],v2.N[VY]),Min(v1.N[VZ],v2.N[VZ]),Min(v1.N[VW],v2.N[VW]));
}

vector4_c max (const vector4_c& v1,const vector4_c& v2)
{
   return vector4_c(Max(v1.N[VX],v2.N[VX]),Max(v1.N[VY],v2.N[VY]),Max(v1.N[VZ],v2.N[VZ]),Max(v1.N[VW],v2.N[VW]));
}

vector4_c operator  - (const vector4_c& v)
{
   return vector4_c(-v.N[VX],-v.N[VY],-v.N[VZ],-v.N[VW]);
}

vector4_c operator  - (const vector4_c& v1, const vector4_c& v2)
{
   return vector4_c(v1.N[VX]-v2.N[VX],v1.N[VY]-v2.N[VY],v1.N[VZ]-v2.N[VZ],v1.N[VW]-v2.N[VW]);
}

vector4_c operator  + (const vector4_c& v1, const vector4_c& v2)
{
   return vector4_c(v1.N[VX]+v2.N[VX],v1.N[VY]+v2.N[VY],v1.N[VZ]+v2.N[VZ],v1.N[VW]+v2.N[VW]);
}

vector_t  operator  * (const vector4_c& v1, const vector4_c& v2)
{
   return v1.N[VX]*v2.N[VX]+v1.N[VY]*v2.N[VY]+v1.N[VZ]*v2.N[VZ]+v1.N[VW]*v2.N[VW];
}

vector4_c operator  * (const vector4_c& v, const vector_t d)
{
   return d*v;
}

vector4_c operator  * (const vector_t d  , const vector4_c& v)
{
   return vector4_c(d*v.N[VX],d*v.N[VY],d*v.N[VZ],d*v.N[VW]);
}

vector4_c operator  * (const vector4_c& v, matrix4_c& m)
{
   return m.transpose()*v;
}

vector4_c operator  * (const matrix4_c& m, const vector4_c& v)
{
   #define ROWCOL(i) m.V[i].N[0]*v.N[VX] + m.V[i].N[1]*v.N[VY] + \
                     m.V[i].N[2]*v.N[VZ] + m.V[i].N[3]*v.N[VW]
   return vector4_c(ROWCOL(0),ROWCOL(1),ROWCOL(2),ROWCOL(3));
   #undef ROWCOL
}

vector4_c operator  / (const vector4_c& v, const vector_t d)
{
   vector_t i = 1.0/d;
   return vector4_c(i*v.N[VX],i*v.N[VY],i*v.N[VZ],i*v.N[VW]);
}

vector4_c operator ^ (const vector4_c& v1, const vector4_c& v2)
{
   return vector3_c
            (
               v1.N[VY]*v2.N[VZ] - v1.N[VZ]*v2.N[VY],
               v1.N[VZ]*v2.N[VX] - v1.N[VX]*v2.N[VZ],
               v1.N[VX]*v2.N[VY] - v1.N[VY]*v2.N[VX]
            );
}

bool operator == (const vector4_c& v1, const vector4_c& v2)
{
   return (v1.N[VX]==v2.N[VX]) && (v1.N[VY]==v2.N[VY]) && (v1.N[VZ]==v2.N[VZ]) && (v1.N[VW]==v2.N[VW]);
}

bool operator != (const vector4_c& v1, const vector4_c& v2)
{
   return !(v1==v2);
}

ostream& operator << (ostream& s, const vector4_c& v)
{
    return s << "(" << v.N[VX] << ", " << v.N[VY] << ", " << v.N[VZ] << ", " << v.N[VW] << ")";
}